/***************************************************************************
 $RCSfile$
 -------------------
 cvs         : $Id: ctapi.c 18 2005-11-29 21:12:26Z aquamaniac $
 begin       : Mon Mar 01 2004
 copyright   : (C) 2004 by Martin Preuss
 email       : martin@libchipcard.de

 ***************************************************************************
 *          Please see toplevel file COPYING for license details           *
 ***************************************************************************/

#ifdef HAVE_CONFIG_H
# include <config.h>
#endif


#include "ctapi-cherry.h"
#include "context_l.h"
#include "drivers/kbd_handler_cherry.h"

#include <stdio.h>
#include <stdlib.h>
#include <assert.h>
#include <string.h>

#include <sys/ioctl.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <linux/kd.h>
#include <errno.h>
#include <unistd.h>
#include <time.h>


/*#define DEBUG_MSGS */



static CC_CONTEXT *CC__contextList=0;
static char CC__atr[4]= {0xa2, 0x13, 0x10, 0x91};
static KBD_HANDLER *CC__kbd_handler=0;
static unsigned int CC__init_count=0;


CC_CONTEXT *CC__findContext(unsigned short ctn) {
  CC_CONTEXT *ctx;

  ctx=CC__contextList;
  while(ctx) {
    if (CC_Context_GetCtn(ctx)==ctn)
      return ctx;
    ctx=CC_Context_Next(ctx);
  }

  return 0;
}



int CC__readdata(CC_CONTEXT *ctx) {
  if (CC__init_count==0) {
    fprintf(stderr, "CTAPI-CHERRY: Reader not initialized.\n");
    return CC_RV_ERR_INVALID;
  }
  else {
    int rv;
    unsigned char pbuf[512];
    unsigned int plen;

#ifdef DEBUG_MSGS
    fprintf(stderr, "INFO: Reading card data.\n");
#endif
    plen=sizeof(pbuf);
    rv=kbd_handler_get_card(CC__kbd_handler, pbuf, &plen);
#ifdef DEBUG_MSGS
    fprintf(stderr, "INFO: Result of reading card data: %d\n", rv);
#endif
    if (rv==0) {
      CC_Context_SetData(ctx, pbuf, plen);
      return 0;
    }
    else if (rv!=KBD_HANDLER_NO_CARD) {
      return -1;
    }

#ifdef DEBUG_MSGS
    fprintf(stderr, "INFO: No card data, handling.\n");
#endif

    rv=kbd_handler_handle(CC__kbd_handler);
#ifdef DEBUG_MSGS
    fprintf(stderr, "INFO: Result of reading card data: %d\n", rv);
#endif
    if (rv==0) {
#ifdef DEBUG_MSGS
      fprintf(stderr, "INFO: Reading card data.\n");
#endif
      plen=sizeof(pbuf);
      rv=kbd_handler_get_card(CC__kbd_handler, pbuf, &plen);
#ifdef DEBUG_MSGS
      fprintf(stderr, "INFO: Result of reading card data: %d\n", rv);
#endif
      if (rv==0) {
#ifdef DEBUG_MSGS
        fprintf(stderr, "INFO: Got card data\n");
#endif
        CC_Context_SetData(ctx, pbuf, plen);
        return 0;
      }
      else if (rv==KBD_HANDLER_NO_CARD) {
#ifdef DEBUG_MSGS
        fprintf(stderr, "INFO: Got no card data\n");
#endif
        return 1;
      }
      else {
#ifdef DEBUG_MSGS
        fprintf(stderr, "ERROR: Could not read card data\n");
#endif
        return 1;
      }
    }
    else if (rv==KBD_HANDLER_NO_CARD)
      return 1;
    else
      return -1;
  }
}



char CT_init(unsigned short ctn, unsigned short pn){
  CC_CONTEXT *ctx;
  int rv;

  fprintf(stderr, "CTAPI-CHERRY: CT_init(%d, %d)\n",
          ctn, pn);

  ctx=CC__findContext(ctn);
  if (ctx) {
    fprintf(stderr, "CTAPI-CHERRY: Context %d already in use\n", ctn);
    return CC_RV_ERR_INVALID;
  }

  /* initialize kbd_handler if necessary */
  if (CC__init_count==0) {
    CC__kbd_handler=kbd_handler_cherry_new();
    rv=kbd_handler_init(CC__kbd_handler);
    if (rv) {
      CC__kbd_handler=0;
      return CC_RV_ERR_CT;
    }
  }
  CC__init_count++;

  ctx=CC_Context_new(ctn, pn);
  assert(ctx);

#ifdef DEBUG_MSGS
  fprintf(stderr, "INFO: Trying to read data...\n");
#endif
  rv=CC__readdata(ctx);
  if (rv<0) {
    CC__init_count--;
    if (CC__init_count==0) {
      kbd_handler_close(CC__kbd_handler);
      kbd_handler_free(CC__kbd_handler);
      CC__kbd_handler=0;
    }
    CC_Context_free(ctx);
    return CC_RV_ERR_TRANS;
  }

  CC_Context_Add(&CC__contextList, ctx);

  fprintf(stderr, "INFO: CT_init(%d, %d): Ok.\n", ctn, pn);
  return CC_RV_OK;
}



int CT__write_atr(CC_CONTEXT *ctx,
                  int p2,
                  unsigned char *p,
                  unsigned short *lenr) {
  int j=0;

  /* copy ATR if wanted */
  if ((p2 & 0xf)==0x00) {
    /* no response */
  }
  else if ((p2 & 0xf)==0x01) {
    int i;

    /* return full ATR */
    if (*lenr<(2+sizeof(CC__atr))) {
      fprintf(stderr,
              "CTAPI_CHERRY: Response buffer too small "
              "(%d, need at least %d bytes)\n",
              *lenr, 2+sizeof(CC__atr));
      return -1;
    }
    for (i=0; i<sizeof(CC__atr); i++) {
      *(p++)=CC__atr[i];
      j++;
    }
  }
  else if ((p2 & 0xf)==0x02) {
    int i;

    /* only return historic bytes */
    if (*lenr<6) {
      fprintf(stderr,
              "CTAPI_CHERRY: Response buffer too small "
              "(%d, need at least %d bytes)\n",
              *lenr, 6);
      return -1;
    }
    for (i=0; i<4; i++) {
      *(p++)=CC__atr[i];
      j++;
    }
  }
  else {
    fprintf(stderr, "CTAPI_CHERRY: Bad value for p2: %d", p2);
    return -1;
  }

  return j;
}



char CT_data(unsigned short ctn,
             unsigned char *dad,
             unsigned char *sad,
             unsigned short lenc,
             unsigned char *command,
             unsigned short *lenr,
             unsigned char *response){
  CC_CONTEXT *ctx;
  unsigned char *p;
  unsigned char c;
  int j=0;
  int rv;

  ctx=CC__findContext(ctn);
  if (ctx==0) {
    fprintf(stderr, "CTAPI-CHERRY: Context %d not in use\n", ctn);
    return CC_RV_ERR_INVALID;
  }

  /* generic parameter checks */
  if (!dad || !sad || !command || !lenr || !response) {
    fprintf(stderr,
            "CTAPI_CHERRY: Null pointer given\n");
    return CC_RV_ERR_INVALID;
  }

  if (*lenr<2) {
    fprintf(stderr,
            "CTAPI_CHERRY: Response buffer too small "
            "(%d, need at least 2 bytes)\n",
            *lenr);
    return CC_RV_ERR_INVALID;
  }

  if (lenc<4) {
    fprintf(stderr,
            "CTAPI_CHERRY: Too few bytes in APDU "
            "(%d, need at least 4 bytes)\n", lenc);
    return CC_RV_ERR_INVALID;
  }

  /*fprintf(stderr, "INFO: Executing command %02x %02x %02x %02x\n",
   command[0], command[1], command[2], command[3]); */
  if (command[0]==0x20) {
    /* terminal commands */
    switch(command[1]) {
    case 0x11: /* RESET ICC */
#ifdef DEBUG_MSGS
      fprintf(stderr, "INFO: Resetting card.\n");
#endif
      if (command[2]!=0x00 && command[2]!=0x01) {
        fprintf(stderr,
                "CTAPI_CHERRY: Only CT or one slot supported (%d)\n",
                command[2]);
        return CC_RV_ERR_INVALID;
      }

      c=*sad;
      *sad=*dad;
      *dad=c;
      p=response;

      CC_Context_SetData(ctx, 0, 0);
      CC_Context_SetReportAbsence(ctx, 1);
      *(p++)=0x62;
      *(p++)=0x00;
      j+=2;
      break;

    case 0x12: /* REQUEST ICC */
      if (command[2]!=0x00 && command[2]!=0x01) {
        fprintf(stderr,
                "CTAPI_CHERRY: Only CT or one slot supported (%d)\n",
                command[2]);
        return CC_RV_ERR_INVALID;
      }

      c=*sad;
      *sad=*dad;
      *dad=c;
      p=response;

      /* fprintf(stderr, "INFO: Getting card.\n"); */
      if (CC_Context_GetReportAbsence(ctx)==1) {
#ifdef DEBUG_MSGS
        fprintf(stderr, "INFO: Reporting absence of the card (req_icc).\n");
#endif
        *(p++)=0x62;
        *(p++)=0x00;
        j+=2;
      }
      else {
        if (CC_Context_GetDataLen(ctx)==0) {
          rv=CC__readdata(ctx);
          if (rv<0)
            return CC_RV_ERR_TRANS;
        }

        if (CC_Context_GetDataLen(ctx)!=0) {
          j=CT__write_atr(ctx, command[3], p, lenr);
          if (j==-1)
            return CC_RV_ERR_INVALID;
          p+=j;
          *(p++)=0x90;
          *(p++)=0x00;
          j+=2;
        } /* if havedata */
        else {
          *(p++)=0x62;
          *(p++)=0x00;
          j+=2;
        }
      }
      break;

    case 0x13: { /* GET STATUS */
      int haveCard=0;

      c=*sad;
      *sad=*dad;
      *dad=c;
      p=response;

      if (CC_Context_GetReportAbsence(ctx)==1) {
#ifdef DEBUG_MSGS
        fprintf(stderr, "INFO: Reporting absence of the card.\n");
#endif
        CC_Context_SetReportAbsence(ctx, 0);
        haveCard=0;
      }
      else {
        if (CC_Context_GetDataLen(ctx)==0) {
          /* fprintf(stderr, "DEBUG: Checking handler\n"); */
          rv=kbd_handler_handle(CC__kbd_handler);
          if (rv<0)
            return CC_RV_ERR_TRANS;
          else if (rv==0) {
            /* fprintf(stderr, "DEBUG: Reading data\n"); */
            rv=CC__readdata(ctx);
            if (rv<0)
              return CC_RV_ERR_TRANS;
          }
          haveCard=(CC_Context_GetDataLen(ctx)!=0);
        }
        else {
          /* fprintf(stderr, "DEBUG: Assuming data is available\n"); */
          haveCard=1;
        }

      }
      if (command[3]==0x80) {
        /* get card status */
        if (*lenr<5) {
          fprintf(stderr,
                  "CTAPI_CHERRY: Response buffer too small "
                  "(%d, need at least %d bytes)\n",
                  *lenr, 5);
          return CC_RV_ERR_INVALID;
        }
        *(p++)=0x80;
        *(p++)=0x01;
	if (haveCard)
          *(p++)=0x05;
        else
          *(p++)=0x00;
        *(p++)=0x90;
        *(p++)=0x00;
        j+=5;
      }
      else if (command[3]==0x46) {
        /* manufacturer info */
        /* TODO */
        fprintf(stderr,
                "CTAPI_CHERRY: Manufacturer info not yet supported\n");
        *(p++)=0x6d; /* not supported */
        *(p++)=0x00;
        j+=2;
      }
      break;
    }

    case 0x14: /* DEACTIVATE ICC */
    case 0x15: /* EJECT ICC */
      c=*sad;
      *sad=*dad;
      *dad=c;
      p=response;
      *(p++)=0x90; /* just return "ok" */
      *(p++)=0x00;
      j+=2;
      CC_Context_SetData(ctx, 0, 0);
      CC_Context_SetReportAbsence(ctx, 1);
      break;

    case 0x18: /* Perform Verification */
    case 0x19: /* Modify Verification Data */
    default:
      c=*sad;
      *sad=*dad;
      *dad=c;
      p=response;
      *(p++)=0x6d; /* not supported */
      *(p++)=0x00;
      j+=2;
      break;
    }

    *lenr=j;
    return CC_RV_OK;
  } /* if terminal command */
  else {
    if (command[0]==0x00 &&
        command[1]==0xb0) {
      int offset;
      int len;
      int i;
      unsigned int dlen;
      const unsigned char *dptr;

      c=*sad;
      *sad=*dad;
      *dad=c;
      p=response;

      if (CC_Context_GetDataLen(ctx)==0) {
        rv=CC__readdata(ctx);
        if (rv<0)
          return CC_RV_ERR_TRANS;
      }

      dlen=CC_Context_GetDataLen(ctx);
      if (dlen!=0) {
        offset=(command[2]<<8)+command[3];
        if (lenc>4) {
          len=command[4];
          if (len==0)
            len=dlen;
        }
        else
          len=dlen;
        if (len>256)
          len=256;
        if (offset>=dlen) {
          fprintf(stderr,
                  "CTAPI_CHERRY: Data out of range\n");
          return CC_RV_ERR_INVALID;
        }
        if (len>(dlen-offset))
          len=dlen-offset;

        if (*lenr<len+2)
          len=*lenr-2;

        /* copy data */
        dptr=CC_Context_GetDataPtr(ctx);
        assert(dptr);
        for (i=0; i<len; i++) {
          *(p++)=dptr[offset+i];
          j++;
        }

        *(p++)=0x90; /* done */
        *(p++)=0x00;
        j+=2;
      } /* if havedata */
      else {
        *(p++)=0x62; /* no card */
        *(p++)=0x00;
        j+=2;
      }
    }
    else if (command[0]==0x00 &&
             command[1]==0xa4) {
      if ((lenc==7 ||  lenc==8) && command[4]==2) {
        int ef;

        c=*sad;
        *sad=*dad;
        *dad=c;
        p=response;

        ef=(command[5]<<8)+command[6];
        if (ef==0x3f00) {
          *(p++)=0x90; /* done */
          *(p++)=0x00;
          j+=2;
        }
        else {
          fprintf(stderr,
                  "CTAPI_CHERRY: SelectFile only works for MF\n");
          *(p++)=0x6a; /* file not found */
          *(p++)=0x82;
          j+=2;
        }
      }
      else {
        fprintf(stderr,
                "CTAPI_CHERRY: SelectFile needs an argument\n");
        return CC_RV_ERR_INVALID;
      }
    }
    else {
      c=*sad;
      *sad=*dad;
      *dad=c;
      p=response;
      *(p++)=0x6d; /* not supported */
      *(p++)=0x00;
      j+=2;
    }

    *lenr=j;
    return CC_RV_OK;
  }

}



char CT_close(unsigned short ctn){
  CC_CONTEXT *ctx;

  if (CC__init_count==0) {
    fprintf(stderr, "CTAPI-CHERRY: No context open.\n");
    return CC_RV_ERR_INVALID;
  }

  ctx=CC__findContext(ctn);
  if (ctx==0) {
    fprintf(stderr, "CTAPI-CHERRY: Context %d not in use\n", ctn);
    return CC_RV_ERR_INVALID;
  }

  CC_Context_Del(&CC__contextList, ctx);
  CC_Context_free(ctx);

  CC__init_count--;
  if (CC__init_count==0) {
    kbd_handler_close(CC__kbd_handler);
    CC__kbd_handler=0;
  }

  return CC_RV_OK;
}






