/***************************************************************************
 $RCSfile$
 -------------------
 cvs         : $Id: kbd_handler_cherry.c 17 2005-11-29 18:26:56Z aquamaniac $
 begin       : Mon Mar 01 2004
 copyright   : (C) 2004 by Martin Preuss
 email       : martin@libchipcard.de

 ***************************************************************************
 *          Please see toplevel file COPYING for license details           *
 ***************************************************************************/

#ifdef HAVE_CONFIG_H
# include <config.h>
#endif

#include "kbd_handler_cherry_p.h"

#include <stdio.h>
#include <errno.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <sys/ioctl.h>
#include <fcntl.h>
#include <assert.h>

//#define DEBUG_MSGS


#define KBD_HANDLER_CHERRY_DATA_OFFSET 0x1e
#define KBD_HANDLER_CHERRY_DEV "/dev/cherry"


KBD_HANDLER *kbd_handler_cherry_new() {
  KBD_HANDLER *kh;
  KBD_HANDLER_CHERRY *ki;

  kh=kbd_handler_new();
  ki=(KBD_HANDLER_CHERRY*)malloc(sizeof(KBD_HANDLER_CHERRY));
  assert(ki);
  memset(ki, 0, sizeof(KBD_HANDLER_CHERRY));

  ki->fd=-1;

  kbd_handler_set_init_fn(kh, kbd_handler_cherry_init);
  kbd_handler_set_close_fn(kh, kbd_handler_cherry_close);
  kbd_handler_set_handle_fn(kh, kbd_handler_cherry_handle);
  kbd_handler_set_get_card_fn(kh, kbd_handler_cherry_get_card);

  kbd_handler_set_private_data(kh, ki, kbd_handler_cherry_free);
  return kh;
}



void kbd_handler_cherry_free(KBD_HANDLER *kh, void *p) {
  if (p) {
    KBD_HANDLER_CHERRY *ki;

    ki=(KBD_HANDLER_CHERRY*)p;
    free((void*)ki);
  }
}



int kbd_handler_cherry_init(KBD_HANDLER *kh) {
  KBD_HANDLER_CHERRY *ki;

  ki=(KBD_HANDLER_CHERRY*)kbd_handler_get_private_data(kh);
  assert(ki);

  ki->fd=open(KBD_HANDLER_CHERRY_DEV, O_RDONLY | O_NONBLOCK);
  if (ki->fd==-1) {
    fprintf(stderr, "ERROR: Could not open device: %s\n", strerror(errno));
    return -1;
  }

  return KBD_HANDLER_OK;
}



void kbd_handler_cherry_close(KBD_HANDLER *kh) {
  KBD_HANDLER_CHERRY *ki;

  ki=(KBD_HANDLER_CHERRY*)kbd_handler_get_private_data(kh);
  assert(ki);

  if (ki->fd!=-1) {
    close(ki->fd);
    ki->fd=-1;
  }
}



int kbd_handler_cherry_handle(KBD_HANDLER *kh) {
  KBD_HANDLER_CHERRY *ki;

  assert(kh);
  ki=(KBD_HANDLER_CHERRY*)kbd_handler_get_private_data(kh);
  assert(ki);

  if (ki->bufsize!=0)
    return KBD_HANDLER_OK;

  if (ki->fd==-1)
    ki->fd=open(KBD_HANDLER_CHERRY_DEV, O_RDONLY | O_NONBLOCK);
  if (ki->fd==-1) {
    fprintf(stderr, "ERROR: Could not open device: %s\n", strerror(errno));
    return -1;
  }

  if (ki->fd!=-1) {
    size_t cnt;
    int rv;

    rv=kbd_handler_input_wait(kh, 500, ki->fd);
    if (rv==-1) {
      close(ki->fd);
      ki->fd=-1;
      //fprintf(stderr, "Timeout.\n");
      return KBD_HANDLER_NO_CARD;
    }
    else if (rv==1) {
      //fprintf(stderr, "No data yet.\n");
      return KBD_HANDLER_NO_CARD;
    }
    cnt=read(ki->fd, ki->buffer, KBD_HANDLER_MAXDATA);
    if (cnt==0) {
      /* reopen and retry if the last read returned EOF */
      close(ki->fd);
      ki->fd=open(KBD_HANDLER_CHERRY_DEV, O_RDONLY | O_NONBLOCK);
      if (ki->fd==-1) {
        fprintf(stderr, "ERROR: Could not open device: %s\n", strerror(errno));
        return -1;
      }
      /* wait for data to be become valid */
      rv=kbd_handler_input_wait(kh, 500, ki->fd);
      if (rv==-1) {
        close(ki->fd);
        ki->fd=-1;
        //fprintf(stderr, "Timeout.\n");
        return KBD_HANDLER_NO_CARD;
      }
      else if (rv==1) {
        //fprintf(stderr, "No data yet.\n");
        return KBD_HANDLER_NO_CARD;
      }
      cnt=read(ki->fd, ki->buffer, KBD_HANDLER_MAXDATA);
      if (cnt==0) {
        fprintf(stderr, "ERROR: Server always returns empty data\n");
        sleep(2);
        close(ki->fd);
        ki->fd=-1;
        return KBD_HANDLER_NO_CARD;
      }
    }

    if (cnt==(size_t)-1) {
      /* error? */
      if (errno==EAGAIN) {
        /*fprintf(stderr, "CTAPI-CHERRY: Nothing to read\n");*/
        /* no data yet, leave device open */
        return KBD_HANDLER_NO_CARD;
      }
      fprintf(stderr, "CTAPI-CHERRY: Could not read (%s)\n",
              strerror(errno));
      close(ki->fd);
      ki->fd=-1;
      return KBD_HANDLER_NO_CARD;
    }
    else {
      /* there is some data */
      ki->bufsize=cnt;
      return KBD_HANDLER_OK;
    }
  }
  return KBD_HANDLER_NO_CARD;
}



int kbd_handler_cherry_get_card(KBD_HANDLER *kh,
                                unsigned char *buffer,
                                unsigned int *bufsize) {
  KBD_HANDLER_CHERRY *ki;
  int rv;

  assert(kh);
  ki=(KBD_HANDLER_CHERRY*)kbd_handler_get_private_data(kh);
  assert(ki);

  if (ki->bufsize) {
    rv=kbd_handler_make_card_data(kh,
                                  ki->buffer,
                                  ki->bufsize,
                                  buffer, bufsize);
    if (rv)
      return rv;
    ki->bufsize=0;
    return KBD_HANDLER_OK;
  }
  else
    return KBD_HANDLER_NO_CARD;
}





